/* Class cBoard
 * Class which represents a Kalah Board
 * does all the neccessary stuff (moves, eval, genMoves...) (values carefully tweaked)
 * implements equals() for Hastable lookup, which returns symmetric boards in regard to P as equal
 * (without looking at Kalah)
 * since Board is symmetric, cBoard always uses numbers from 0-6 and uses delta() to get actual bowl for current P
 * also has different eval functions for testing, EVAL_FUNC selects which to use
*/
import java.util.*;

public class cBoard{
    
	public int[] balls = new int[14];
	public byte iNext=0;
	
	protected static int NUMSTONES 		= 6;
	protected static int WEIGH_MKMOVES	= 2;
	protected static int WEIGH_MQMOVES	= 5;
	protected static int WEIGH_MBOWLS	= 2;
	protected static int WEIGH_MKALAH	= 5;
	protected static int WEIGH_OKMOVES	= 2;
	protected static int WEIGH_OQMOVES	= 5;
	protected static int WEIGH_OBOWLS	= 2;
	protected static int WEIGH_OKALAH	= 5;
	protected static int OFFSET		= 0;	
	protected static int EVAL_FUNC		= 6;
		
	public int uBound,lBound,bestMove=-1,worstMove=-1,depth=-1;
	
    
    public cBoard(){
		Arrays.fill(balls,NUMSTONES);
		balls[6]=0;
		balls[13]=0;	
    }
    public cBoard(int k, int b, int km, int qm){
		this();
		WEIGH_MBOWLS  = b;
		WEIGH_MKALAH  = k;
		WEIGH_MKMOVES = km;
		WEIGH_MQMOVES = qm;
	}
    public cBoard(int eval, int mk, int mb, int mkm, int mqm,int ok, int ob, int okm, int oqm){
		this();
		EVAL_FUNC     = eval; 
		WEIGH_MBOWLS  = mb;
		WEIGH_MKALAH  = mk;
		WEIGH_MKMOVES = mkm;
		WEIGH_MQMOVES = mqm;
		WEIGH_OBOWLS  = ob;
		WEIGH_OKALAH  = ok;
		WEIGH_OKMOVES = okm;
		WEIGH_OQMOVES = oqm;
    }
	
    /* returns the true weigh for this chosen val */
    private int w(int w){
	if (w<13) return w;			//0 , 6 or 12
	else if (w==13) return (int)( Math.floor(getBallsInGame()/6) ); 		//start game - var
	else if (w==14) return (int)( 12-Math.floor(getBallsInGame()/6) ); 	//end game - var
    	else return 6;
    }
    /* returns true iff the game is still alive */
    public boolean isAlive(){
		return !isEmpty();
    }
    /* returns all the balls as an array */
    public int[] getState(){
    	return balls;
    }
    /* returns the balls in one bowl */
    public int getBowl(int iNum){
    	return balls[iNum];
    }
    /* returns the balls in one bowl-delta version*/
    public int getBowlD(int iNum){
    	return balls[d(iNum)];
    }
    /* returns the next player to move (0,1) */
    public byte getNext(){
	return iNext;
    }
    public int getBallsInGame(){
    	int ret=0;
	for(int i=0;i<13;i++)
	    if (i!=6) ret+=balls[i];
	return ret;
    }
    
    /* makes a move */
    public void move(int move)
    {
 	if (move > 6) move-=7;
	//cannot move from Kalah
	if (move==6) return;
 	gamelogic(move);
 	uBound=Integer.MAX_VALUE;lBound=Integer.MIN_VALUE;
	bestMove=worstMove=-1;
    }
    
    /* Equals method... for the Hashtable to work with the lookup
    return wether two board states are equal without taking into credit the kalahs
    since the board is symmetric in respect to both players, a turned table with the other
    player to move is also equal (delta function makes sure of that) */
    public boolean equals(Object o){
    	boolean E=true;
	if (o instanceof cBoard){
		for (int h=0; h<14; h++)
			if ( (h!=6 && h!=13) &&
				balls[d(h)] != ((cBoard)o).balls[((cBoard)o).d(h)]) { E=false; break; }	 
    	}else E=false;	
    	return E;
    }
    
    public String toString(){
    	String ret = "";
	for (int i=0;i<14;i++) ret +=( ((ret!="")?",":"") + balls[i]);
    	return ("Board{A="+isAlive()+",s="+iNext+",e="+(eval())+",{"+ret+
		"},u="+uBound+",l="+lBound+",b="+bestMove+",w="+worstMove+"}");
    }
        
    /* eval position */
    public int eval(){
    	return this.eval(iNext);
    }
    public int eval(int n){
    	switch(EVAL_FUNC){
		case 1: return eval1(n);
		case 2: return eval2(n);
		case 3: return eval3(n);
		case 4: return eval4(n);
		case 5: return eval5(n);
		case 6: return ultimateEval(n);
	}
	return eval1(n);
    }
    /*---------------*/
    
    /* gives the number of possible Kalah moves, i.e. moves which end in Kalah, for this player */
    private int km(int side){
    	int k=0,i;
    	for (i=0+side*7;i<6+side*7;i++)
		if (balls[i]==6-i) k++;			// Kalah move
	return k;
    }
    /* gives the number of possible Quiescence moves, i.e. moves which end in one's empty bowl and move Opponents
    stones to one's Kalah, for this player */
    private int qm(int side){
    	int q=0,i;
    	for (i=0+side*7;i<6+side*7;i++){
			int eb = (i+balls[i])%13;			//ending bowl
				if (  (balls[i]>0 )		&&		//balls in bowl
		   			(eb<(6+side*7))		&&		//end up in one of my bowls
		   			(balls[eb]==0 || eb==i)	&& 		//ending bowl empty or same bowl
		   			( balls[ (12-eb) ] > 0 ) ) q++;		//opposite bowl not empty: qmove
		}
		return q;
    }		
    /* like qm, but gives the overall number of Opponents Stones one is able to move */
    private int q(int side){
    	int q=0,i,j;
		for (i=0+side*7;i<6+side*7;i++){
			int eb = (i+balls[i])%13;			//ending bowl
			if (  (balls[i]>0 )		&&		//balls in bowl
		   		(eb<(6+side*7))		&&		//end up in one of my bowls
		   		(balls[eb]==0 || eb==i)	&& 		//ending bowl empty or same bowl
		   		( (j=balls[ (12-eb) ]) > 0 ) ) q+=j;		//opposite bowl not empty
		}
		return q;
    }
    /* gives the number of stones in player's bowls */
    private int bowls(int side){
		int k=0,i;
		for (i=0+side*7;i<6+side*7;i++)
			k += balls[i];
		return k;
    }

    
    //eval1: just stones in kalah... works coz # is linear and dependent on stones in other kalah
    private int eval1(int side){
    	return ( balls[(side==0)?6:13] );
    }
    //eval2: difference in stones in kalah
    private int eval2(int side){
    	return ( balls[(side==0)?6:13]-balls[(side==0)?13:6] );
    }
    //eval3: stones in kalah + some info about advantage of current pos (km & qm)
    private int eval3(int side){
    	return	( eval2(side)
			+ (km(side) * WEIGH_MKMOVES )
			+ (qm(side) * WEIGH_MQMOVES )
		);
	 
    }
    //eval4: stones in kalah + some more info about advantage of current pos (km& q)
    //i.e. the number of stones a Q move would get
    private int eval4(int side){
    	return	( eval2(side)
			+ (km(side) * WEIGH_MKMOVES )
			+ (q(side)  * WEIGH_MQMOVES )
		);
    }
    
    private int eval5(int my){
    	int op = (my+1)%2;
	return ( OFFSET						//a little offset
		+ ( balls[(my==0)?6:13] * w(WEIGH_MKALAH) ) 	//balls in my kalah
		- ( balls[(my==0)?13:6] * w(WEIGH_MKALAH) ) 	//balls in his kalah
		+ ( bowls(my) 		* w(WEIGH_MBOWLS) ) 	//balls in my bowls
		- ( bowls(op) 		* w(WEIGH_MBOWLS) )	//balls in his bowls
		+ ( km(my) 		* w(WEIGH_MKMOVES) )	//number of my kalah moves
		- ( km(op) 		* w(WEIGH_MKMOVES) )	//number of his kalah moves
		+ ( q(my) 		* w(WEIGH_MQMOVES) )	//stones movable by Q move
		- ( q(op) 		* w(WEIGH_MQMOVES) )	//stones movable by Q move
	);
    }
    //UltimateEvil: Everything
    private int ultimateEval(int my){
    	int op = (my+1)%2;
	return ( OFFSET							//a little offset
			+ ( balls[(my==0)?6:13] * w(WEIGH_MKALAH) ) 	//balls in my kalah
			- ( balls[(my==0)?13:6] * w(WEIGH_OKALAH) ) 	//balls in his kalah
			+ ( bowls(my) 		* w(WEIGH_MBOWLS) ) 	//balls in my bowls
			- ( bowls(op) 		* w(WEIGH_OBOWLS) )	//balls in his bowls
			+ ( km(my) 		* w(WEIGH_MKMOVES) )	//number of my kalah moves
			- ( km(op) 		* w(WEIGH_OKMOVES) )	//number of his kalah moves
			+ ( q(my) 		* w(WEIGH_MQMOVES) )	//stones movable by Q move
			- ( q(op) 		* w(WEIGH_OQMOVES) )	//stones movable by Q move
		);
   }    
    
    
    
    /* ------End eval thingies ------*/
    public void copy(cBoard b){
    	System.arraycopy(b.balls,0,balls,0,14);
	iNext=b.iNext;
    }
    
    //isQuiescence: checks whether current state alows a zero bowl move 
    public boolean isQuiescence(){
    	boolean E=true;
	for (int i=0+iNext*7;i<6+iNext*7;i++){
		int eb = (i+balls[i])%13;			//ending bowl
		if (  (balls[i]>0 )		&&		//balls in bowl
		   (eb<6+iNext*7)		&&		//end up in one of my bowls
		   (balls[eb]==0 || eb==i)	&& 		//ending bowl empty or same bowl
		   ( balls[ (12-eb) ] > 0 ) )	//opposite bowl not empty:
		 	{ E=false;break; }		//means wild swing in eval...
	}
    	return E;
    }
    
    
    //delta function: gives the correct bowl for current player
    public int d(int b){ return (b+iNext*7)%14; }
    
    //private thingies...
    
    //isEmpty: determines if current player ended game...
    private boolean isEmpty(){
    	boolean M=false,H=false;
	int h;
	//all my bowls empty
	for (h=0; h<6; h++)
		if (balls[d(h)]!=0) { M=true; break; }
	//or all his bowls empty
	for (h=7; h<13; h++)
		if (balls[d(h)]!=0) { H=true; break; }
	return (!M || !H);
    }
    //gamelogic: implements the gamelogic
    private void gamelogic(int current){
	int i,last;
	int chips=balls[d(current)];
	balls[d(current)]=0;
	
	last=playerMove(current,chips);
	
	if (isEmpty()){  		//finalize scores if ended
	    for (i=0;i<6;i++)		//take my bowls
	    {
		    balls[d(6)] += balls[d(i)];
		    balls[d(i)] =  0;
	    }
	    for (i=7;i<13;i++)		//take other players bowls
	    {
		    balls[d(13)] += balls[d(i)];
		    balls[d(i)] =  0;
	    }
	    String msg;
	    if (balls[6]==balls[13]) msg="Nobody won; 36:36";
	    else msg="Player"+((balls[6]>balls[13])?"1":"2") +" won; "+
	    	((balls[6]>balls[13])?balls[6]:balls[13]) + ":" + ((balls[6]>balls[13])?balls[13]:balls[6]);
	    cLog.Log(3,"Board","Game ended for this board, "+msg);
	    return;
	}
	
	if (last != 6){ //if player gets a ball in his kalah he goes again
	    iNext = (byte) (((iNext+1))%2);
	}
    }
    //playermove: calls itself with the next bowl to put chips into (always assumes player 0 (bowls 0-6),
    //d() makes the rest
    private int playerMove(int current,int chips){
	if (chips==0) return 0;
	current = (current+1)%13;  //move to the next bowl before adding balls, and dont put in opps. kalah
	if (chips==1){
	    if ( (balls[d(current)]==0) && (current < 6)){ //put last ball in empty cup get contents of oposite cup
		    int opisitbowl = 12-(d(current));
		    balls[d(6)] += balls[opisitbowl]+1;
		    balls[opisitbowl]=0;
		    balls[d(current)]=0;
		    return(current);
	    }
	    balls[d(current)]++;
	    return(current);		    
	}
	balls[d(current)]++;
	return(playerMove(current,--chips));
	
    }
    //class which represents a move in what GenMove returns
    public class cMove{
    	public int e=0;			//extend search depth by
    	public int m=-1;			//the actual move
	public cMove(int pm){ m=pm; }
	public cMove(int pm, int ex){ m=pm; e=ex; }
    }
    
    public Vector genMoves(int side){
	int i;
	cLog.Log(5,"Board","Gen'ing moves for "+this.toString());
	Vector vs = new Vector(6);
	Vector vd = new Vector(6);
	for (i=0;i<6;i++) vs.add(new Integer(i));
	int bm = ( (getNext()==side)?bestMove:worstMove );
	//old best move
	if (bm>-1){ 
		vs.remove(new Integer(bm));
		vd.add(   new cMove(bm,1));		//setting this to 2 makes it loose by 2	
	} 
	
	
	//do kalah moves , also from far side first, coz this doesnt fuck up remaining kalah moves...
	//kalah moves first is clearly better.. makes it win
	for (i=vs.size()-1;i>=0;i--){
		int m = ((Integer)vs.get(i)).intValue();	//get the move as int
		if (balls[d(m)]==6-m){ 				// Kalah move
 			vs.remove(i);
			vd.add(new cMove(m,1));			//should be 1
		}
	}
	
	//then do moves which end up in my empty bowl
	//including this moves is also better...
	//but doing it first, makes it loose
	i=0;
	while (i<vs.size()){
		int m = ((Integer)vs.get(i)).intValue();	//get the move as int
		int eb = (m+balls[d(m)])%13;			//ending bowl
		if (  (balls[d(m)]>0 )	&&			//balls in bowl
	   		(eb<6)		&&			//end up in one of my bowls
	   		(balls[ d(eb) ]==0 || eb==m)	&& 	//ending bowl empty or same bowl
	   		( balls[ (12-d(eb)) ] > 0 ) ){		//opposite bowl not empty:
				vs.remove(0);
				vd.add(new cMove(m,0));		//setting this to 1 makes it to look too deep
				
		}else i++;
	}
	
	//add other moves in order
	while (vs.size()>0){
		vd.add(new cMove(((Integer)vs.remove(0)).intValue() ));
	}
	return vd;
   }
    
}
