/* Class cKalah
 * Main class which implements all the stuff for the Window and Commandline
 * uses cMTDF to search and cBoard for the Board representation
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

public class cKalah implements ActionListener
{
	//some consts (have to be static in java)
	/*------------------------------------*/
	private static boolean bWnd=false;
	protected static byte SEARCHDEPTH = 20;
	protected static int  TIMEOUT	= 20;
	protected static int  LOGLEVEL	= -1;
	/*------------------------------------*/
	
	private static cBoard kb;
	private static JButton[] Bowls;
	private static boolean bRand=false;
	private static byte iPlayer=-1;
	private static Random rd;
	private static BufferedReader br;
	private static AIThread AIt;
	private static JButton MoveB;
	private static JTextField TOField;
	private static JTextField LLField;
	private static long startTimeMillis;
	private static int numGensPending;
	private static Vector numOppsPending=new Vector();
	private static int ttcounter;
	private static JFrame jf;
		
	public static void main(String[] args){
		int i;
		cLog.level = LOGLEVEL;
		if (args.length<1){
			cLog.Log(-1,"init","Please enter player number PC will play");
			System.exit(0);
		}
		for (i=0;i<args.length;i++){
			try{ iPlayer = (byte)(( (Byte.parseByte(args[i]))==1)?1:0 ); }
			catch(Exception ex){}
		}
		if (iPlayer==-1){
			cLog.Log(-1,"init","Error parsing player number");
			System.exit(1);
		}
			
		
		if (args.length>1 && args[1].equals("rand")){
			bRand=true;
			cLog.Log(1,"init","Randomizing PC moves");
		}
		Arrays.sort(args);
		if (Arrays.binarySearch(args,"-w")>=0)
			bWnd=true;
		
		if (iPlayer==1){
// 			p0(7,13,7,8,20,17,9,13,6,8)=32
			SEARCHDEPTH = 20;
			TIMEOUT	= 17;
			kb = new cBoard(6,7,13,7,8,9,13,6,8);
		}else{
// 			p1(12,13,11,13,20,20,12,13,12,13)=38
// 			p1(12,13,11,13,15,15,12,13,12,13)=34
			SEARCHDEPTH = 20;
			TIMEOUT	= 20;
			kb = new cBoard(6,12,13,11,13,12,13,12,13);
		}
		
		br = new BufferedReader(new InputStreamReader(System.in),5);
		
		cKalah k = new cKalah();
		
	}
	
	public cKalah(){
		long time=System.currentTimeMillis();
		rd = new Random();
		
		if (bWnd){
			cLog.level = LOGLEVEL = 1;
			makeWnd();
			//and start background timer
			java.util.Timer t = new java.util.Timer(true);
			t.schedule(new tt(),0,100);
		
		}else{			//go into infinite loop calling the timerfunc every 100ms
			tt t = new tt();
			try{
				while(true){
					if ( (System.currentTimeMillis()-time)>100){
						t.run();	
						time=System.currentTimeMillis();
					}
				}
			}catch(Exception e){ e.printStackTrace(); System.err.println("Error while sleeping"); }
		}
	}

	public void actionPerformed(ActionEvent e){
		int i;
		byte n = kb.getNext();
		
		if (e.getActionCommand().equals("Gen")){
			if (AIt==null){
					//search
					cLog.Log(1,"Gen","Generating move for P"+(n+1)+",e="+kb.eval(n));
					startTimeMillis = System.currentTimeMillis();
					AIt = new AIThread(kb);
					AIt.start();
					if(bWnd){ MoveB.setText("Gen("+TOField.getText()+")");MoveB.setEnabled(false);}
					
				
			}else numGensPending++;
		}else if(e.getActionCommand().equals("stop")){
			if (AIt!=null){
				AIt.bRunning=false;
				AIt.stop();
				try{
					AIt.join();
				}catch(InterruptedException ex){
					cLog.Log(0,"Gen","Error in waiting for search thread:");
					ex.printStackTrace();
				}
				int bm = kb.bestMove;
				int fg = AIt.firstguess;
				AIt=null;
				if( ( (n==0) && (bm<0) && (bm>5)) || ( (n==1) && (bm<7) && (bm>12) ) ){
				//search again
					System.err.println("Bad move,fg="+fg+",bm="+bm);
					cLog.Log(1,"Gen","Generating move for P"+(n+1)+",e="+kb.eval(n));
					startTimeMillis = System.currentTimeMillis();
					AIt = new AIThread(kb,fg);
					AIt.start();
					if(bWnd){ MoveB.setText("Gen("+TOField.getText()+")");MoveB.setEnabled(false);}
				}
					
				if(bWnd){ MoveB.setText("Gen");MoveB.setEnabled(true);}
				move(bm);
				
				//output all pos moves, with corr. AB(f) vals: cMTDF.MoveTab
				//doesnt really seem to output correct values
				for (i=0;i<6;i++){
					if (cMTDF.MoveTab[i]>Integer.MIN_VALUE)
						cLog.Log(1,"Gen","Possible move from Bowl"+(i+n*7)+" with"
						+ "AB="+cMTDF.MoveTab[i]);
				}
				cLog.Log(1,"Gen","MTDF chose bowl"+ (bm+n*7) + " with fg="+fg+",\n\tnew "+kb);
				cLog.Log(-1,"",""+(bm+n*7));
				i=0;
				//and delete some entries from the HT
				for (Iterator it = cMTDF.ht.values().iterator() ; it.hasNext() ; ) {
         				cBoard b = (cBoard)it.next();
					//and remove if more than current balls in game
					if (b.getBallsInGame()>kb.getBallsInGame()){
						it.remove();
						i++;
					}
				}
				cLog.Log(1,"MTDF",i+" Bs deleted, size:"+cMTDF.ht.size());
     			}
		}else if(e.getActionCommand().equals("LLField")){
			cLog.level = Integer.parseInt(LLField.getText());
			cLog.Log(0,"Log","new Level set");
		}else{	//move
			if (AIt!=null) numOppsPending.add(e.getActionCommand());
			else{
				try{
					move(Integer.parseInt(e.getActionCommand()));
				}catch (Exception ex){ex.printStackTrace();}
			}
		}
	}
	
	protected void move(int m){
		int i;
		kb.move(m);
		//update board and alive
		if(bWnd){
			int state[] = kb.getState();
			for (i=0;i<14;i++)
				Bowls[i].setText(new Integer(state[i]).toString());
			jf.setTitle("cKalah v0.9 - turn P"+(kb.getNext()+1));
		}
		if (!kb.isAlive()){
			int[] balls=kb.getState();
			if(bWnd){
				String t="";
				for (i=0;i<14;i++) Bowls[i].setEnabled(false);
				if (balls[6]==balls[13]) t="draw";
				else if(balls[6]>balls[13]) t="P1 won";
				else t="P2 won";
				jf.setTitle("Kalah v0.9 - "+t);
			}
			
			String msg;
			if (balls[6]==balls[13]) msg="Nobody won; 36:36";
			else msg="Player"+((balls[6]>balls[13])?"1":"2") +" won; "+
			   ((balls[6]>balls[13])?balls[6]:balls[13]) + ":" + ((balls[6]>balls[13])?balls[13]:balls[6]);
			cLog.Log(1,"Kalah","Game over, "+msg);
		}
		if(bWnd) jf.repaint();
	}
	
	//private class for timer
	private class tt extends TimerTask{
		public void run(){
// 			System.err.println("TTcount:"+ (++ttcounter));
			
			//Moves pending?
			if (AIt==null && numOppsPending.size()>0 ){
				actionPerformed(new ActionEvent(this,0,(String)numOppsPending.remove(0)));
			}
			//Gens pending?
			if (AIt==null &&  (numGensPending--)>0 ){
				actionPerformed(new ActionEvent(this,0,"Gen"));
			//check if there's something to be read
			}
			try{
					if (br.ready()){
						String input = br.readLine();
								
						if (input.equals("quit")) System.exit(0);
						else if (input.equals("stop"))
							actionPerformed(new ActionEvent(this,0,"stop"));
						else if (input.equals("move"))
							actionPerformed(new ActionEvent(this,0,"Gen"));
						else if (input.startsWith("opponent"))
							actionPerformed(new ActionEvent(this,0,input.substring(9)));
						
					}
				}catch(IOException e){ System.err.println("Error while reading from in"); }
			
			System.err.flush();
			//and check if theres a thinking thread which time is up
			try{
				if (AIt!=null){
					long to,run = System.currentTimeMillis() - startTimeMillis;
					if (bWnd){ MoveB.setText("Gen("+(run/100)+")");
						to = Integer.parseInt(TOField.getText());
					}else to = TIMEOUT;
					if ( ((run/100)>to) || !AIt.isAlive() )
						actionPerformed(new ActionEvent(this,0,"stop"));
				}else{ //check if button isnt reactivated
					if (bWnd && !MoveB.isEnabled()) MoveB.setEnabled(true);
				}
			}catch (Exception e){
				e.printStackTrace();
				System.err.println("Error while stopping AIt");
			}
		}
	}
	
	/* Private class for AI thread
	   is started for each move... */
	private class AIThread extends Thread{
		public volatile boolean bRunning=true;
		public cBoard kb;
		public volatile int firstguess=-100;
		
		public AIThread(cBoard c){
			kb=c;
		}
		public AIThread(cBoard c, int fg){
			kb=c;firstguess=fg;
		}
		public void run(){
			if (!kb.isAlive()) return;
			cLog.Log(3,"AIThread","MTDF started");
			if (firstguess==-100)
				firstguess = kb.eval()+1;//take current eval + 1 as first guess for real AB(f)
			Arrays.fill(cMTDF.MoveTab,Integer.MIN_VALUE);	//and fill Move Table with lower Bound
			try{
				//do iterative deeping until time is over
				for (int i=1;i<=SEARCHDEPTH;i++){
					cLog.Log(1,"AIThread","MTDF("+i+"):"+firstguess+",move:"+kb.bestMove);
					firstguess = cMTDF.MTDF(kb,firstguess,i);
					if (!bRunning) break;	
				}
				
			}catch(Exception e){ 
				e.printStackTrace();
			}
			finally{
				bRunning=false;
			}
		}
	}
	
	private void makeWnd(){
		jf = new JFrame("cKalah v0.9");
		jf.setDefaultCloseOperation(jf.EXIT_ON_CLOSE);
		
		JPanel jp1 = new JPanel();
		jp1.setLayout(new BorderLayout(5,5));
		jp1.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
		jf.setSize(580,200);
		jf.getContentPane().add(jp1);
		
		Bowls = new JButton[14];
		
		//add two kalahs
		JButton jb = new JButton();
		jb.setEnabled(false);
		Bowls[(iPlayer==0)?6:13]=jb;
		jp1.add(jb,(iPlayer==0)?BorderLayout.EAST:BorderLayout.WEST);
		
		jb = new JButton();
		jb.setEnabled(false);
		Bowls[(iPlayer==0)?13:6]=jb;
		jp1.add(jb,(iPlayer==1)?BorderLayout.EAST:BorderLayout.WEST);
		
		//make GenMove Pane
		JPanel jp2 = new JPanel(new GridLayout(1,6,2,2));
		MoveB = new JButton("Gen");
		MoveB.setActionCommand("Gen");
		MoveB.addActionListener(this);
		jp2.add(MoveB);
		
		//make timeout thingie
		TOField = new JTextField(new Integer(TIMEOUT).toString());
		JLabel jl = new JLabel("sec TO");
		jp2.add(TOField);
		jp2.add(jl);
		
		jb = new JButton("stop");
		jb.setActionCommand("stop");
		jb.addActionListener(this);
		jp2.add(jb);
		
		//make loglevel thingie
		LLField = new JTextField(new Integer(LOGLEVEL).toString());
		LLField.setActionCommand("LLField");
		LLField.addActionListener(this);
		jl = new JLabel("LogLevel");
		jp2.add(LLField);
		jp2.add(jl);
		
		jp1.add(jp2,BorderLayout.NORTH);	
		
		//make panel and add bowls
		int i;
		JPanel jp = new JPanel();
		jp.setLayout(new GridLayout(2,6,20,20));
		
		//make bowls
		for (i=12;i>6;i--){
			jb = new JButton();
			jb.setActionCommand(new Integer(i).toString());
			jb.addActionListener(this);
			if (iPlayer==0) jb.setEnabled(false);
			Bowls[i]=jb;
			jp.add(jb);
			
		}
		for (i=0;i<6;i++){
			jb = new JButton();
			jb.setActionCommand(new Integer(i).toString());
			jb.addActionListener(this);
			if (iPlayer==1) jb.setEnabled(false);
			Bowls[i]=jb;
			jp.add(jb);
		}
		jp1.add(jp,BorderLayout.CENTER);
		
		//update board and alive
		int state[] = kb.getState();
		for (i=0;i<14;i++)
			Bowls[i].setText(new Integer(state[i]).toString());
		jf.setTitle("cKalah v0.9 - turn P"+(kb.getNext()+1));
		jf.setVisible(true);
	}
}
